/** @example ex/Intro/QueueBrowser.c
 */

/*
 *  Copyright 2007-2020 Solace Corporation. All rights reserved.
 *
 *  http://www.solace.com
 *
 *  This source is distributed under the terms and conditions
 *  of any contract or contracts between Solace and you or
 *  your company. If there are no contracts in place use of
 *  this source is not authorized. No support is provided and
 *  no distribution, sharing with others or re-use of this
 *  source is authorized unless specifically stated in the
 *  contracts referred to above.
 *
 *  QueueBrowser
 *
 *  This sample shows how an external user can
 *  retrieve the latest orderbook snapshot from Solace.
 *  It makes use of a Solace API Browser Flow
 *  https://docs.solace.com/Solace-PubSub-Messaging-APIs/API-Developer-Guide/Browsing-Guaranteed-Mess.htm
 *  
 *  
 *  Since the queue that the client connects to is an LVQ,
 *  there will only be a single message available (latest orderbook snapshot message)
 *  
 *  This is a sample for demonstration purposes only
 * 
 *  Parameters:
 *      -c, --cip=[Protocol:]Host[:Port] Protocol, host and port of the Solace messaging appliance (e.g. --cip=tcp:192.168.160.101).
 *	    -u, --cu=user[@vpn] Client username and Mesage VPN name.  
 *	    -q, --queue=Queue   Queue Name. 
 *	    -p, --cp=password   Client password. 
 *   Where OPTIONS are:
 *  	-l, --log=loglevel  API and application logging level (debug, info, notice, warn, error, critical).
 * 
 */

#include "os.h"
#include "solclient/solClient.h"
#include "solclient/solClientMsg.h"
#include "common.h"

/**
* @struct rxBrowseFlowCallbackInfo
*
* A structure that is used by the browser flow Rx callback 
* this example shows that it can be used as a counter for # received messages
*/
typedef struct rxBrowseFlowCallbackInfo {
    solClient_uint32_t      msgCount;          /**< Number of received messages. */
} rxBrowseFlowCallbackInfo_t;

/*****************************************************************************
 * flowMessageReceiveCallback
 *
 * The message receive callback for the session.
 * This callback is triggered asynchronously when a message is received from the endpoint 
 * that the consumer is bound to
 * 
 *****************************************************************************/
static          solClient_rxMsgCallback_returnCode_t
flowMessageReceiveCallback ( solClient_opaqueFlow_pt opaqueFlow_p, solClient_opaqueMsg_pt msg_p, void *user_p )
{
    solClient_msgId_t msgId;
    void           *binaryAttachmentBuffer_p;
    solClient_uint32_t binaryAttachmentBufferSize;
    rxBrowseFlowCallbackInfo_t      *callbackInfo_p;
    solClient_returnCode_t          rc;

    if (user_p == NULL) {
        /* Indicate test suite failure */
        solClient_log(SOLCLIENT_LOG_ERROR, "Got null user pointer when reading message");
        return SOLCLIENT_CALLBACK_OK;
    }
    callbackInfo_p = (rxBrowseFlowCallbackInfo_t *)user_p;    

    if (opaqueFlow_p == NULL) {
        solClient_log(SOLCLIENT_LOG_ERROR, "Got null opaque flow pointer when reading message");
        return SOLCLIENT_CALLBACK_OK; /* not taking msg */ 
    }

    if (msg_p == NULL) {
        solClient_log(SOLCLIENT_LOG_ERROR, "Got null message pointer when reading message");
        return SOLCLIENT_CALLBACK_OK; /* not taking msg */ 
    }

    /* Process the message. */
    // example of how you can retrieve the messageID from a message

    rc = solClient_msg_getMsgId(msg_p, &msgId);
    if (rc  != SOLCLIENT_OK) {
            solClient_log(SOLCLIENT_LOG_WARNING, "Unable to retrieve the message ID when reading message");
        return SOLCLIENT_CALLBACK_OK; 
    }

    // example of retrieving user payload and printing it to stdout
    if ( ( rc = solClient_msg_getBinaryAttachmentPtr ( msg_p,
                                                       &binaryAttachmentBuffer_p,
                                                       &binaryAttachmentBufferSize ) ) != SOLCLIENT_OK ) {
        common_handleError ( rc, "solClient_msg_getBinaryAttachmentPtr()" );
        return SOLCLIENT_CALLBACK_OK;
    }
    // refer to the Solace C API reference for other accessor methods 

    callbackInfo_p->msgCount++;

    // As an example, lets dump the entire messge to stdout
    printf ( "Received message on browser flow with MsgID = %lld. Message Dump:\n", msgId );
    printf ( "<======================>\n" );
    solClient_msg_dump ( msg_p, NULL, 0 );
    printf ( "<======================>\n" );
    printf ( "\n" );

    return SOLCLIENT_CALLBACK_OK;
}

/*****************************************************************************
 * browserFlow
 *
 * Wrapper function that creates a browser flow session,
 * binds to a given queue,
 * and browses a single orderbook snapshot message
 * 
 *****************************************************************************/
static void 
browserFlow(solClient_opaqueSession_pt   session_p,
            char                         *queueName,
            solClient_uint32_t           browseWindow)
{   
    solClient_opaqueFlow_pt             flow_p = NULL;
    const char                         *flowProps[20];
    int                                 propIndex;
    solClient_flow_createFuncInfo_t     flowFuncInfo    =
        SOLCLIENT_SESSION_CREATEFUNC_INITIALIZER;
    rxBrowseFlowCallbackInfo_t          flowCallbackInfo;
    
    char                                browseWindowStr[32];

    snprintf(browseWindowStr, sizeof(browseWindowStr), "%d", browseWindow);

    flowCallbackInfo.msgCount = 0;

    /*
     * Set Callback function for flow
     */

    flowFuncInfo.rxMsgInfo.callback_p = flowMessageReceiveCallback;
    flowFuncInfo.rxMsgInfo.user_p     = &flowCallbackInfo;
    flowFuncInfo.eventInfo.callback_p = common_flowEventCallback;

    /*
     * Set Consumer Flow Properties
     */

    propIndex = 0;
    flowProps[propIndex++] = SOLCLIENT_FLOW_PROP_BIND_BLOCKING;
    flowProps[propIndex++] = SOLCLIENT_PROP_ENABLE_VAL;

    flowProps[propIndex++] = SOLCLIENT_FLOW_PROP_BIND_ENTITY_ID;
    flowProps[propIndex++] = SOLCLIENT_FLOW_PROP_BIND_ENTITY_QUEUE;

    flowProps[propIndex++] = SOLCLIENT_FLOW_PROP_BIND_NAME;
    flowProps[propIndex++] = queueName;

    /*
     * Set 'browser' mode and window size
     */
    flowProps[propIndex++] = SOLCLIENT_FLOW_PROP_BROWSER;
    flowProps[propIndex++] = SOLCLIENT_PROP_ENABLE_VAL;
    flowProps[propIndex++] = SOLCLIENT_FLOW_PROP_WINDOWSIZE;
    flowProps[propIndex++] = browseWindowStr;

    flowProps[propIndex] = NULL;

    if (solClient_session_createFlow( (char **) flowProps,
                                     session_p,
                                     &flow_p,
                                     &flowFuncInfo,
                                     sizeof(flowFuncInfo)) != SOLCLIENT_OK) { 
        solClient_log(SOLCLIENT_LOG_ERROR, "solClient_session_createFlow() did not return SOLCLIENT_OK");
        
        goto cleanup;
    }

    printf("Browsing messages from queue '%s' .....\n", queueName);

    /*
     * Wait for one message to be received, since the queue will
     * only contain a single orderbook snapshot message
     */

    while ( flowCallbackInfo.msgCount < 1 ) {
        SLEEP ( 1 );
    }

    // After one message is received, destroy the browser flow
    
 cleanup: 
    if (flow_p != NULL) {
        if (solClient_flow_destroy(&flow_p) != SOLCLIENT_OK) {
            solClient_log(SOLCLIENT_LOG_ERROR, "Could not destroy browser flow");
            
        }
    }
}

/*****************************************************************************
 * main
 *
 * The entry point to the application.
 *****************************************************************************/
int
main ( int argc, char *argv[] )
{
    struct commonOptions                commandOpts;
    solClient_returnCode_t              rc = SOLCLIENT_OK;
    
    /* Context */
    solClient_opaqueContext_pt context_p;
    solClient_context_createFuncInfo_t contextFuncInfo = SOLCLIENT_CONTEXT_CREATEFUNC_INITIALIZER;

    /* Session */
    solClient_opaqueSession_pt session_p;
    solClient_session_createFuncInfo_t sessionFuncInfo = SOLCLIENT_SESSION_CREATEFUNC_INITIALIZER;

    /* Flow */
    solClient_opaqueFlow_pt flow_p;
    solClient_flow_createFuncInfo_t flowFuncInfo = SOLCLIENT_FLOW_CREATEFUNC_INITIALIZER;

    /* 
        Flow Properties 
        We only want the browser to receive
        a single message
    */
    solClient_uint32_t                  browseWindow = 1;

    printf("\nQueueBrowser.c (Copyright 2007-2020 Solace Corporation. All rights reserved.)\n\n");

    /*************************************************************************
    * Parse command options
    *************************************************************************/
    common_initCommandOptions(&commandOpts, 
                               ( USER_PARAM_MASK |
                                 HOST_PARAM_MASK |
                                 PASS_PARAM_MASK |
                                 QUEUE_PARAM_MASK ),    /* required parameters */
                               ( 
                                LOG_LEVEL_MASK ));    /* optional parameters */
    if (common_parseCommandOptions(argc, argv, &commandOpts, NULL) == 0) {
        exit(1);
    }

    /*************************************************************************
    * Optional - uncomment the below lines to redirect all API logs to a file
    * 'solace.log'
    *************************************************************************/

    /*
    const char  solaceLogFileStr[] = "solace.log";

    if ((rc = solClient_log_setFile(solaceLogFileStr)) != SOLCLIENT_OK) {
        common_handleError(rc, "solClient_log_setFile");
        goto notInitialized;
    }
    */

    /*************************************************************************
     * Initialize the API and setup logging level
     *************************************************************************/

    /* solClient needs to be initialized before any other API calls. */
    if ((rc = solClient_initialize(SOLCLIENT_LOG_DEFAULT_FILTER,
                                   NULL)) != SOLCLIENT_OK) {
        common_handleError(rc, "solClient_initialize()");
        goto notInitialized;
    }

    common_printCCSMPversion();

    /* 
     * Standard logging levels can be set independently for the API and the
     * application. In this case, the ALL category is used to set the log level for 
     * both at the same time.
     */
    solClient_log_setFilterLevel(SOLCLIENT_LOG_CATEGORY_ALL,
                                 commandOpts.logLevel);

    /*************************************************************************
     * Create a Context
     * 
     * Contexts are used for organizing communication between an application and 
     * a Solace PubSub+ event broker.
     * 
     * A Context encapsulates threads that drive network I/O and 
     * message delivery notification for the Sessions and Session components
     * associated with that Context.
     *************************************************************************/

    /* 
     * Create a Context, and specify that the Context thread be created
     * automatically instead of having the application create its own
     * Context thread.
     */
    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient context");

    if ((rc = solClient_context_create(SOLCLIENT_CONTEXT_PROPS_DEFAULT_WITH_CREATE_THREAD,
                                       &context_p,
                                       &contextFuncInfo,
                                       sizeof(contextFuncInfo))) != SOLCLIENT_OK) {
        common_handleError(rc, "solClient_context_create()");
        goto cleanup;
    }

    /*************************************************************************
     * Create and connect a Session
     * 
     * A Session creates a single TCP client connection to a Solace event broker 
     * for sending and receiving messages.
     *************************************************************************/

    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient session");

    if ((rc = common_createAndConnectSession(context_p,
                                             &session_p,
                                             common_messageReceivePrintMsgCallback,
                                             common_eventCallback,
                                             NULL,
                                             &commandOpts)) != SOLCLIENT_OK) {
        common_handleError(rc, "common_createAndConnectSession()");
        goto cleanup;
    }

    if (!solClient_session_isCapable(session_p, SOLCLIENT_SESSION_CAPABILITY_BROWSER)) {
      solClient_log(SOLCLIENT_LOG_ERROR, "Stopping as Solace appliance doesn't have guaranteed delivery enabled");
        goto sessionConnected;
    }

    /*************************************************************************
     * Create a flow to browse messages on the queue 
     * 
     * A subscriber Flow allows applications to receive Guaranteed messages
     * from an endpoint, such as a Queue.
     ***************************************************/
    
    browserFlow (session_p, commandOpts.queueName, browseWindow);

    printf ( "Endpoint Browsing completed. Exiting.\n" );

    /*************************************************************************
     * Cleanup
     *************************************************************************/

  sessionConnected:
    /* Disconnect the Session. */
    solClient_log(SOLCLIENT_LOG_INFO, "Disconnecting solClient session");
    if ( ( rc = solClient_session_disconnect ( session_p ) ) != SOLCLIENT_OK ) {
        common_handleError ( rc, "solClient_session_disconnect()" );
    }

  cleanup:
    /* Cleanup solClient. */
    if ((rc = solClient_cleanup()) != SOLCLIENT_OK) {
        common_handleError(rc, "solClient_cleanup()");
    }
    
  notInitialized:
    return 0;

}
